#!/bin/bash

ARGSPEC="--pm-file:PM_FILE --pctl-file:PCTL_FILE \
--rows:ROWS --cols:COLS --num-teams:NUM_TEAMS \
--first-team:FTM --last-team:LTM \
--team-size:TEAM_SIZE --deadline:DEADLINE"

#check arguments
function usage {
    echo "Usage : $0 <options>"
    echo "Options :"
    for X in $ARGSPEC; do
        X1=`echo $X | awk -F':' '{print $1}'`
        X2=`echo $X | awk -F':' '{print $2}'`
        echo "     $X1 $X2"
    done
    exit 1
}

#parse options
while :
do
    case $1 in
        --pm-file|--pctl-file|--rows|--cols|\
        --num-teams|--first-team|--last-team|--team-size|--deadline)
            for X in $ARGSPEC; do
                X1=`echo $X | awk -F':' '{print $1}'`
                X2=`echo $X | awk -F':' '{print $2}'`
                if [ "$X1" == "$1" ]; then
                    eval ${X2}=$2
                    shift 2
                    break
                fi
            done
            ;;
        [a-zA-Z0-9\-]*)
            echo "ERROR: Unknown option: $1" >&2
            usage
            exit 1
            shift
            ;;
        *)
            break
            ;;
    esac
done

#check all options defined
for X in $ARGSPEC; do
    X1=`echo $X | awk -F':' '{print $1}'`
    X2=`echo $X | awk -F':' '{print $2}'`
    if [ -z "${!X2}" ]; then
        echo "ERROR: ${X1} undefined"
        usage
    fi
    echo "$X2 = ${!X2}"
done

cat <<EOF > $PM_FILE
///////////////////////////////////////////////////////////////////////
//if COORD_ALGO = 0 then:
///////////////////////////////////////////////////////////////////////
//a demining scenario with a single robot moving in a zig zag
//pattern. the robot stops if : (a) it explodes; or (b) it has exceeded
//its deadline; or (c) if it has covered 1/N of the total number of
//cells. in the first two cases, the robot has failed. in the last case,
//the robot has succeeded. we compute the probability of success and
//raise it to the N-th power. this way, we get the probability of
//success if N robots independently demined 1/N of the field. varying N
//gives us different results, and we want the smallest N for which the
//result exceeds 0.8.
///////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////
//if COORD_ALGO = 1 then:
///////////////////////////////////////////////////////////////////////
//a demining scenario with a team of N robots moving in a zig zag
//pattern. the team consists of a leader, and followers maintaining a
//fixed distance from the leader. if the leader explodes, then the
//followers (if any left) elect a new leader. the team stops if : (a)
//the leader explodes and no robots are left; or (b) it has exceeded its
//deadline; or (c) if it has covered all cells. in the first two cases,
//the team has failed. in the last case, the team has succeeded. we
//compute the probability of success. varying N gives us different
//results, and we want the smallest N for which the result exceeds 0.8.
///////////////////////////////////////////////////////////////////////

dtmc

///////////////////////////////////////////////////////////////////////
//parameters supplied by command line
///////////////////////////////////////////////////////////////////////
//detecting capability -- should be 0, 1 or 2
const int DETECTING_CAPABILITY;
//defusing capability -- should be 0, 1 or 2
const int DEFUSING_CAPABILITY;
//locomotion capability -- should be 0, 1 or 2
const int LOCOMOTION_CAPABILITY;

//probability of detecting a mine. we assume no false positives. so if
//a mine is detected, then there is a mine in the cell.
const double P_DETECT_MINE;

//the terrain -- maps detecting capability to probability of a false
//negative, and gives probabilities for being in the three defusing
//situations.
const double P_FN_SC0;
const double P_FN_SC1;
const double P_FN_SC2;
const double P_IN_DEFUSE1;
const double P_IN_DEFUSE2;
const double P_IN_DEFUSE3;

//probabilities of exploding while detecting a mine -- depends on the
//detecting capability
const double P_ED_SC0;
const double P_ED_SC1;
const double P_ED_SC2;

//probabilities of exploding while defusing a mine -- depends on the
//defusing capability
const double P_ED_DC0;
const double P_ED_DC1;
const double P_ED_DC2;

//probability of successful skip -- depends on locomotion capability
const double P_SKIP_LC0;
const double P_SKIP_LC1;
const double P_SKIP_LC2;

//probability of successful turn -- depends on locomotion capability
const double P_TURN_LC0;
const double P_TURN_LC1;
const double P_TURN_LC2;

//probability of success in leader election
const double P_ELECT_LEADER;

///////////////////////////////////////////////////////////////////////
//constants
///////////////////////////////////////////////////////////////////////

//initial and final cell coordinates
EOF

#general initial and final cell coordinates for various teams

#check if cells equally divisible among teams
NCELLS=`expr $ROWS \* $COLS`
RCELLS=`expr $NCELLS % $NUM_TEAMS`
if [ "$RCELLS" != "0" ]; then
    echo "ERROR: cannot divide $NCELLS cells among $NUM_TEAMS teams"
    exit 1
fi

CURRX="0"
CURRY="0"
ROWSM1=`expr $ROWS - 1`
COLSM1=`expr $COLS - 1`
NTM1=`expr $NUM_TEAMS - 1`
CPT=`expr $NCELLS / $NUM_TEAMS`
for i in `seq 0 $NTM1`; do
    IP1=`expr $i + 1`
    A=`expr $i + 1`
    A=`expr $CPT \* $A`
    NEXTX=`expr $A / $ROWS`
    NEXTY=`expr $A % $ROWS`
    E=`expr $NEXTX % 2`
    if [ "$E" == "1" ]; then
        NEXTY=`expr $ROWSM1 - $NEXTY`
    fi
    E=`expr $CURRX % 2`
    if [ "$E" == "1" ]; then
        INITDIR="true"
    else
        INITDIR="false"
    fi

    A=`expr $A - 1`
    FINALX=`expr $A / $ROWS`
    FINALY=`expr $A % $ROWS`
    E=`expr $FINALX % 2`
    if [ "$E" == "1" ]; then
        FINALY=`expr $ROWSM1 - $FINALY`
    fi

    cat <<EOF >> $PM_FILE
const bool INIT_DIR_$IP1 = $INITDIR;
const int INIT_X_$IP1 = $CURRX;
const int INIT_Y_$IP1 = $CURRY;
const int FINAL_X_$IP1 = $FINALX;
const int FINAL_Y_$IP1 = $FINALY;
EOF

    CURRX=$NEXTX
    CURRY=$NEXTY
done

cat <<EOF >> $PM_FILE

//robot states
const int INIT = 0;
const int NEXT = 1;
const int WORK = 2;
const int BLOWNUP = 3;
const int DONE = 4;
const int NOT_DETECTED_MINE = 5;
const int DEFUSE1 = 6;
const int DEFUSE2 = 7;
const int DEFUSE3 = 8;
const int MARK = 9;
const int ELECT_LEADER = 10;
const int STUCK = 11;
const int NEXT_COL1 = 12;
const int NEXT_COL2 = 13;
const int LAST = 13;

//maximum X and Y axis values -- smallest is 0
const int X_MAX = $COLSM1;
const int Y_MAX = $ROWSM1;

//probability of blowing up while detecting a mine -- this depends on
//the detecting capability
const double p_explode_detect = (DETECTING_CAPABILITY=0) ? P_ED_SC0 :
                                (DETECTING_CAPABILITY=1) ? P_ED_SC1 : P_ED_SC2;

//probabilty of false negative -- this depends on the detecting capability
const double p_false_neg = (DETECTING_CAPABILITY=0) ? P_FN_SC0 : 
                           (DETECTING_CAPABILITY=1) ? P_FN_SC1 : P_FN_SC2;

//probabilities of blowing up in each defusing state. this depends on
//the defusing capability.
const double p_explode_defuse1 = (DEFUSING_CAPABILITY=0) ? P_ED_DC0 : 
                                 (DEFUSING_CAPABILITY=1) ? P_ED_DC1 : P_ED_DC2;
const double p_explode_defuse2 = (DEFUSING_CAPABILITY=0) ? P_ED_DC0 : 
                                 (DEFUSING_CAPABILITY=1) ? P_ED_DC1 : P_ED_DC2;
const double p_explode_defuse3 = (DEFUSING_CAPABILITY=0) ? P_ED_DC0 : 
                                 (DEFUSING_CAPABILITY=1) ? P_ED_DC1 : P_ED_DC2;
const double p_explode_mark = (DEFUSING_CAPABILITY=0) ? P_ED_DC0 : 
                              (DEFUSING_CAPABILITY=1) ? P_ED_DC1 : P_ED_DC2;

//probability of success in turning and skipping
const double p_skip = (LOCOMOTION_CAPABILITY=0) ? P_SKIP_LC0 : 
                      (LOCOMOTION_CAPABILITY=1) ? P_SKIP_LC1 : P_SKIP_LC2;
const double p_turn = (LOCOMOTION_CAPABILITY=0) ? P_TURN_LC0 : 
                      (LOCOMOTION_CAPABILITY=1) ? P_TURN_LC1 : P_TURN_LC2;

//various times
const int t_sense = 1; //time to sense mine
const int t_defuse1 = 1; //time to defuse in easy case
const int t_defuse2 = 2; //time to defuse in medium case
const int t_defuse3 = 3; //time to defuse in hard case
const int t_mark = 4; //time to defuse in hard case
const int t_elect_leader = 1; //time to elect new leader

///////////////////////////////////////////////////////////////////////
//team ${FTM}
///////////////////////////////////////////////////////////////////////
module Team${FTM}
    //local state
    s${FTM} : [INIT..LAST] init INIT;

    //whether all operations succeeded
    success${FTM} : bool init true;

    //current X and Y coordinate value
    x${FTM} : [0..X_MAX] init INIT_X_${FTM};
    y${FTM} : [0..Y_MAX] init INIT_Y_${FTM};
    dir${FTM} : bool init INIT_DIR_${FTM};

    //team size
    team_size${FTM} : [0..$TEAM_SIZE] init $TEAM_SIZE;

    //after detecting, if no timeout happends, either a mine was
    //detected or not, or the robot blew up. also, we increase the
    //clock by the time it took to sense.
    [tick] (s${FTM}=INIT) -> 
       p_explode_detect:(s${FTM}'=ELECT_LEADER) + 
       P_DETECT_MINE:(s${FTM}'=DEFUSE1) + 
       (1-p_explode_detect-P_DETECT_MINE):(s${FTM}'=NOT_DETECTED_MINE);

    //if mine not detected, blow up or move to next (successfully,
    //i.e., there really was no mine, or with failure, i.e., a mine
    //was there but was not detected). we assume that if there was a
    //mine (i.e., a false negative), then there is a .5 probability of
    //blowing up.
    [tick] s${FTM}=NOT_DETECTED_MINE -> 
       p_false_neg * 0.5:(s${FTM}'=ELECT_LEADER) + 
       p_false_neg * 0.5:(s${FTM}'=NEXT)&(success${FTM}'=false) +
       (1-p_false_neg):(s${FTM}'=NEXT);

    //if defuse1, next, defuse2, or explode
    [tick] (s${FTM}=DEFUSE1) -> 
       (P_IN_DEFUSE1*(1-p_explode_defuse1)):(s${FTM}'=NEXT) + 
       P_IN_DEFUSE1*p_explode_defuse1:(s${FTM}'=ELECT_LEADER) + 
       (1-P_IN_DEFUSE1):(s${FTM}'=DEFUSE2);

    //if defuse2, next, defuse3, or explode
    [tick] (s${FTM}=DEFUSE2) -> 
       (P_IN_DEFUSE2*(1-p_explode_defuse2)):(s${FTM}'=NEXT) + 
       P_IN_DEFUSE2*p_explode_defuse2:(s${FTM}'=ELECT_LEADER) + 
       (1-P_IN_DEFUSE2):(s${FTM}'=DEFUSE3);

    //if defuse3, next, mark, or explode
    [tick] (s${FTM}=DEFUSE3) -> 
       (P_IN_DEFUSE3*(1-p_explode_defuse3)):(s${FTM}'=NEXT) + 
       P_IN_DEFUSE3*p_explode_defuse3:(s${FTM}'=ELECT_LEADER) + 
       (1-P_IN_DEFUSE3):(s${FTM}'=MARK);

    //if mark, next, or explode
    [tick] (s${FTM}=MARK) -> 
       (1-p_explode_mark):(s${FTM}'=NEXT) + 
       p_explode_mark:(s${FTM}'=ELECT_LEADER);

    //leader election: no more robots left in team, blow up
    [tick] (s${FTM}=ELECT_LEADER)&(team_size${FTM}=1) -> (s${FTM}'=BLOWNUP);

    //leader election: if more robots left in team, probabilistically 
    //elect new leader and resume
    [tick] (s${FTM}=ELECT_LEADER)&(team_size${FTM}>1) -> 
       P_ELECT_LEADER:(s${FTM}'=NEXT)&(team_size${FTM}'=team_size${FTM}-1) +
       (1-P_ELECT_LEADER):(s${FTM}'=STUCK)&(team_size${FTM}'=team_size${FTM}-1);

    //if blown up, done or stuck, be stuck
    [tick] s${FTM}=BLOWNUP -> (s${FTM}'=BLOWNUP);
    [tick] s${FTM}=STUCK -> (s${FTM}'=STUCK);
    [tick] s${FTM}=DONE -> (s${FTM}'=DONE);

    //choosing the next cell

    //if you are at the final cell,
    [tick] (s${FTM}=NEXT)&(x${FTM}=FINAL_X_${FTM})&(y${FTM}=FINAL_Y_${FTM}) -> 
      (s${FTM}'=DONE);

    //if you are at the top moving up, move right and then down
    [tick] (s${FTM}=NEXT)&((x${FTM}!=FINAL_X_${FTM})|(y${FTM}!=FINAL_Y_${FTM}))&
           (!dir${FTM})&(y${FTM}=Y_MAX)&(x${FTM}+1<=X_MAX) -> 
      p_turn:(s${FTM}'=NEXT_COL1)&(dir${FTM}'=true)&(x${FTM}'=x${FTM}+1) +
      (1-p_turn):(s${FTM}'=STUCK);

    [tick] (s${FTM}=NEXT_COL1) -> 
      p_skip:(s${FTM}'=NEXT_COL2) + (1-p_skip):(s${FTM}'=STUCK);

    [tick] (s${FTM}=NEXT_COL2) -> 
      p_turn:(s${FTM}'=INIT) + (1-p_turn):(s${FTM}'=STUCK);

    //if you are not at the top moving up, move up
    [tick] (s${FTM}=NEXT)&((x${FTM}!=FINAL_X_${FTM})|(y${FTM}!=FINAL_Y_${FTM}))&
           (!dir${FTM})&(y${FTM}<Y_MAX) -> 
      p_skip:(s${FTM}'=INIT)&(y${FTM}'=y${FTM}+1) + (1-p_skip):(s${FTM}'=STUCK);

    //if you are at the bottom moving down, move right and then up
    [tick] (s${FTM}=NEXT)&((x${FTM}!=FINAL_X_${FTM})|(y${FTM}!=FINAL_Y_${FTM}))&
           (dir${FTM})&(y${FTM}=0)&(x${FTM}+1<=X_MAX) -> 
      p_turn:(s${FTM}'=NEXT_COL1)&(dir${FTM}'=false)&(x${FTM}'=x${FTM}+1) +
      (1-p_turn):(s${FTM}'=STUCK);

    //if you are not at the bottom moving down, move down
    [tick] (s${FTM}=NEXT)&((x${FTM}!=FINAL_X_${FTM})|(y${FTM}!=FINAL_Y_${FTM}))&
           (dir${FTM})&(y${FTM}>0) -> 
      p_skip:(s${FTM}'=INIT)&(y${FTM}'=y${FTM}-1) + (1-p_skip):(s${FTM}'=STUCK);
endmodule
EOF

for i in `seq $FTM $LTM`; do
    if [ "$i" != "$FTM" ]; then
        cat <<EOF >> $PM_FILE

///////////////////////////////////////////////////////////////////////
//team $i
///////////////////////////////////////////////////////////////////////
module Team$i = Team${FTM} [ 
    s${FTM}=s$i, success${FTM}=success$i, 
    x${FTM}=x$i, y${FTM}=y$i, dir${FTM}=dir$i,
    team_size${FTM}=team_size$i,
    INIT_DIR_${FTM}=INIT_DIR_$i,
    INIT_X_${FTM}=INIT_X_$i,INIT_Y_${FTM}=INIT_Y_$i,
    FINAL_X_${FTM}=FINAL_X_$i,FINAL_Y_${FTM}=FINAL_Y_$i
] endmodule
EOF
    fi
done

cat <<EOF >> $PM_FILE

///////////////////////////////////////////////////////////////////////
//reward structure to keep track of number of cells covered
///////////////////////////////////////////////////////////////////////
rewards "covered_cells"
EOF

for i in `seq $FTM $LTM`; do
    cat <<EOF >> $PM_FILE
    (s$i=INIT) : 1.0;
EOF
done

cat <<EOF >> $PM_FILE
endrewards
EOF

SUCC=""
for i in `seq $FTM $LTM`; do
    SUCCI=""
    for j in `seq 0 $DEADLINE`; do
        if [ "$SUCCI" == "" ]; then
            SUCCI="(s${i}=DONE & success${i})"
        else
            SUCCI="(s${i}=DONE & success${i}) | (X$SUCCI)"
        fi
    done


    if [ "$SUCC" == "" ]; then
        SUCC="($SUCCI)"
    else
        SUCC="$SUCC & ($SUCCI)"
    fi
done

cat <<EOF > $PCTL_FILE
"success_prob" : P=? [ $SUCC ]
"expected_covered" : R{"covered_cells"}=? [ C<=$DEADLINE ]
EOF
